/* Main for GXPS converters
 *
 * Copyright (C) 2011  Carlos Garcia Campos <carlosgc@gnome.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <config.h>

#include <glib.h>
#include <locale.h>
#include G_STRINGIFY(CONVERTER_HEADER)

gint main (gint argc, gchar **argv)
{
        GXPSConverter *converter;

        setlocale (LC_ALL, "");

        converter = GXPS_CONVERTER (g_object_new (CONVERTER_TYPE, NULL));

        if (!gxps_converter_init_with_args (converter, &argc, &argv)) {
                g_object_unref (converter);

                return 1;
        }

        gxps_converter_run (converter);
        g_object_unref (converter);

        return 0;
}
